;Liest Daten aus zwei HX711-Chips (2 Wgezellen) und einem Hall-sensor und sendet diese ber UART
;Hall-Daten werden alle 1,024ms verschickt, Messdaten je nach Verfgbarkeit
;Bit 7 in Hall-Byte zeigt an, ob noch Messdaten folgen (Bit7=1)
;;--------------------------------------------------------------------------
;; Connections
;;  Clock an beide HX711 - RC3, Pin7
;;  Daten von 1. HX711 	 - RC4, Pin6
;;  Daten von 2. HX711 	 - RC5, Pin5
;;  Daten von Hallsensor - AN6, Pin8
;;  UART Tx              - RC0, Pin10
;;  UART Rx				 - INT/RA2, Pin11
;;--------------------------------------------------------------------------
;;
;Interner Clock mit 4MHz/4 = 1Mhz
;Timer0 mit Prescale 1:4 erzeugt Sample-Clock: 256*4=1024 Zyklen = 1,024ms
;UART mit Baud-Rate 38,4kBaud

;;Config-register sollte bei Programmierung auf 11 xxx1 1000 0001 (0x3181) stehen.

	processor  16F676
	radix  DEC
	__CONFIG 0x3181

;Register Adressen
INDF        EQU   0x00
TMR0        EQU   0x01
STATUS      EQU   0x03
FSR         EQU   0x04
PORTA       EQU   0x05
PORTC       EQU   0x07
INTCON      EQU   0x0B
PIR1	    EQU   0x0C
TMR1L       EQU   0x0E
TMR1H       EQU   0x0F
T1CON       EQU   0x10
CMCON		EQU   0x19
ADRESH      EQU   0x1E
ADCON0      EQU   0x1F
OPTIONR     EQU   0x81
TRISA       EQU   0x85
TRISC       EQU   0x87
PIE1        EQU   0x8C
PCON        EQU   0x8E
OSCCAL      EQU   0x90
ANSEL       EQU   0x91
WPUA	    EQU   0x95
IOCA	    EQU   0x96
VRCON	    EQU   0x99
EEDAT	    EQU   0x9A
EEADR	    EQU   0x9B
EECON1	    EQU	  0x9C
EECON2      EQU   0x9D
ADRESL      EQU   0x9E
ADCON1      EQU   0x9F

_C      	EQU 	0		;Carry-Bit
_Z      	EQU   	2		;Zero-Bit
_W			EQU		0		;Destination W bei z.B. INCF
_R			EQU		1		;Destination selbes Register bei z.B. INCF 

#define	_BANK0	BCF STATUS,5	;Umschalten auf Bank0
#define	_BANK1	BSF STATUS,5	;Umschalten auf Bank1
#define	_RSTX	PORTC,0			;UART Tx
#define	_RSRX	PORTA,2			;UART Rx (=INT)
#define	_RC1	PORTC,1			;Output zum Debug
#define _ADCGO	ADCON0,1		;Go-Bit fr ADC
#define _GIE	INTCON,7		;Global Interrupt enable
#define _INTF	INTCON,1		;Interrupt-Flag UART-RX
#define _T0IF	INTCON,2		;Interrupt-Flag TMR0
#define _CARRY	STATUS,0		;CArry-Bit in Status
#define _ZERO	STATUS,2		;Zero-Bit in Status
#define	_CLK	PORTC,3			;Clock an beide HX711
#define	_DAT1	PORTC,4			;Daten von 1. HX711
#define	_DAT2	PORTC,5			;Daten von 2. HX711
#define	_RUN	flags,0			;Run-Flag

;Memory
mem			EQU 	0x20
m			EQU		mem				;Variablen fr UART
n			EQU		mem+1	
temp		EQU		mem+2
tdat1		EQU		mem+3			;Temporres Lesebyte fr Dat1
tdat2		EQU		mem+4			;Temporres Lesebyte fr Dat2
D1H			EQU		mem+5			;Die finalen 24-Bit Daten (High/Mid/Low)- in dieser Reihenfolge belassen, werden durch FSR beschrieben
D2H			EQU		mem+6
D1M			EQU		mem+7
D2M			EQU		mem+8
D1L			EQU		mem+9
D2L			EQU		mem+10
udat		EQU		mem+11			;Zwischendatum im UART-Empfang
flags		EQU		mem+12			;Flags: Bit 0 = Run
DPos		EQU		mem+13
scnt		EQU		mem+14			;Sende-Count: Zeit an wieviele Messbyte-Double noch versendet werden mssen
INTWSAVE	EQU	  	mem+62
INTSSAVE	EQU	  	mem+63


;******* Programm Start **************************************************
	ORG 0
		CLRF  	PORTA			
		CLRF  	PORTC			;Setzt UART-TX auf low, d.h. Zustand vor Startbit, Clk an HX711 auf low
 		GOTO	main


;******* Interrupt Service Routine ***************************************
	ORG	4
;Mgliche Interrupt-Quellen und ihre Funktion
;- INT steigende Flanke: UART-Empfang vom PC starten
ISR		BCF		_GIE			;Alle Interrupts disablen
		MOVWF	INTWSAVE		;preserve w register
		SWAPF	STATUS,_W		;preserve status register
		MOVWF	INTSSAVE
ISRchk	;Prfung, was den Int ausgelst hat
		BTFSS	_INTF			;INT steigende Flanke?
		GOTO	ISRend			;Nein: ISR ignorieren
		BSF		_RC1			;Ja: Debug-Puls ausgeben
		BCF		_RC1
;		BTFSS	_RSRX			;Ist Pin immer noch High?
;		GOTO	ISRend			;Nein: Dann war der Empfang wohl nur ein fehlerhafter Spike
		CALL	RSrx			;UART lesen
		MOVWF	udat
		BSF		_RC1			;Zweiten Debug-Puls ausgeben
		BCF		_RC1			
		MOVF	udat,_R			;Empfangenes Byte prfen
		BTFSC	_ZERO			;=0 (Startbit und 8 Datenbits=5V)?		
		GOTO	$+3				;Ja: Start-Kommando
		BCF		_RUN			;Nein: Alles andere ist Stopp-Kommando, geht leider nicht anders ISR ist lange whrernd einem Tx gesperrt, d.h. der Empfang eines spezifischen Bytes kann nicht sichergestellt werden. Jedes Startbit wird mehr oder weniger als Stop-Kommando gewertet
		GOTO	ISRend			
		BSF		_RUN			;Start-Kommando	
ISRend	BCF		_INTF			;Flag zurcksetzen - erst nach dem RSrx, weil ansonsten ein Bitwechsel in der bertragung das Flag wieder setzen und einen weiteren Interrupt auslsen wrde
		SWAPF	INTSSAVE,_W		;restore STATUS register
		MOVWF	STATUS
		SWAPF	INTWSAVE,_R		;restore w register
		SWAPF	INTWSAVE,_W
		RETFIE					;enabled gleichzeitig auch wieder alle Interrupts



;******** Main *************************************************
main	;Prozessor Init
		MOVLW	0x10
		MOVWF 	INTCON		;GIE noch disabled, INTE enabled	
		MOVLW	0x00
		MOVWF 	T1CON		;TMR1:disabled
		MOVLW	0x07
		MOVWF 	CMCON		;Comparator:disabled
		MOVLW 	0x19		;ADC: Left justified, VRef=Vdd, Channel=6, ADC an aber nocht nicht aktiv
		MOVWF 	ADCON0
		_BANK1				;Umschalten auf Bank 1
		MOVLW 	0xC1		;Options: INT auf steigende Flanke (Startbit), TMR0 prescale=1:4		
		MOVWF 	OPTIONR
		MOVLW 	0x0C		
		MOVWF 	TRISA		;PortA: Bis auf RA2(UART-RX) und RA3(kann nur Input) alles Outputs
		MOVLW	0x34
		MOVWF 	TRISC		;PortC: Bis auf C2(ADC-In), C4 und C5(HX711 Inputs) alles Output
		CLRF  	PIE1		;Keine Peripherie-Interruptrs
		MOVLW	0x40
		MOVWF 	ANSEL		;Nur AN6 ist ADC-Pin
		CLRF	WPUA		;Keine Pull-ups
		CLRF	IOCA		;Keine Interrupt-on change auf PortA (INT ist separat)
		CLRF	VRCON		;Voltage Reference off
		MOVLW 	0x10
		MOVWF 	ADCON1		;ADCON1: 8 Tosc bei 4MHz = 2us Conversion Zeit
		_BANK0 				;Bank 0
		BCF		_RUN
		BSF		_GIE		;Interrupts enablen	

start	BTFSS	_RUN		;Wartet bis Run=1 (durch UART-Rx in ISR)	
		GOTO	$-1
		CLRF	TMR0		;1ms-Timer zurcksetzen
		BCF		_T0IF
rnlp	BTFSS	_RUN		;Run-Loop, solange bis _RUN=0 (durch UART-Rx in ISR)
		GOTO	start		;RUN=0: Run-Loop beenden, auf neuen Start warten
		BTFSS	_T0IF		;Warte bis 1ms-Timer auslst
		GOTO	rnlp
		;RUN ist noch 1 und 1ms-Timer hat ausgelst: Mess- und bertragungszyklus starten
		BCF		_T0IF
		BCF		_GIE		;Whrend der Messung/bertragung bitte keine Interrupts - die knnten das Transmit-Timing vernichten
		BSF   	_ADCGO		;Starte ADC Conversion - Acquisition-Time muss nicht beachtet werden, da AD-Kanal nicht ungeschaltet wird
		BTFSC 	_ADCGO		;Pruefe, ob Konvertierung beendet (GO=0)
		GOTO  	$-1			;Wenn nein, weiterwarten
		MOVF  	ADRESH,_W	;Wenn ja, ADRESH...	
		MOVWF 	DPos		;...nach DPos speichern
		RRF		DPos,_R		;DPos(7) soll das Signalbit enthalten, ob Messdaten folgen oder nicht
		;Prfen, ob HX711 beide neue Daten beit liegen haben (beide DAT=low)
		BTFSC	_DAT1		;Ist Dat1=low (ready vom HX711)?
		GOTO	SPos		;Nein: Nur Positions-Byte senden
		BTFSC	_DAT2		;Ja: Ist Dat2=low (ready vom HX711)?
		GOTO	SPos		;Nein: nur Positions-Byte senden
		MOVLW	D1H			;Ja: HX711-Daten einholen, Schreibpointer vorbereiten
		MOVWF	FSR
		MOVLW	.3			;3 Lese-Loops in denen jeweils 8 Bit von Dat1 und 8 Bit von Dat2 gelesen werden
		MOVWF	n			;Gesamtzeit dafr: 441us
lpbyt	MOVLW	.8
		MOVWF	m
lpbit	BSF		_CLK		;Clk-High: HX711 legen neues Bit auf Output - Zeit zw. Pulsen: 17us
		BCF		_CLK
		BTFSC	_DAT1		;Pruefe Bit von DAT1
		GOTO	$+3			;=1
		BCF		_CARRY		;=0
		GOTO	$+3
		BSF		_CARRY		;Carry mit 0 besetzen
		NOP					;Damit gleich lang wie BCF-Zweig
		RLF		tdat1,_R	;Carry in Temp-Byte verschieben (MSB first)
		BTFSC	_DAT2		;Pruefe Bit von DAT2
		GOTO	$+3			;=1
		BCF		_CARRY		;=0
		GOTO	$+3
		BSF		_CARRY		;Carry mit 0 besetzen
		NOP					;Damit gleich lang wie BCF-Zweig
		RLF		tdat2,_R	;Carry in Temp-Byte verschieben (MSB first)
		DECFSZ	m,_R
		GOTO	lpbit		;Bit-Loop 8 mal wiederholen
		;2x 8 Bit wurden gelesen und stehen in tdat1 und tdat2
		MOVF	tdat1,_W	;Bytes in die jeweiligen Endpositionen schreiben
		MOVWF	INDF
		INCF	FSR,_R
		MOVF	tdat2,_W
		MOVWF	INDF
		INCF	FSR,_R
		DECFSZ	n,_R
		GOTO	lpbyt		;Byte-Loop 3 mal wiederholen
		BSF		_CLK		;HX711 erwartet noch einen 25. Puls zum Fortfahren mit Gain=128
		BCF		_CLK
;		BSF		_CLK		;Bzw. noch einen 26. und 27. Puls zum Fortfahren mit Gain=64
;		BCF		_CLK
;		BSF		_CLK		
;		BCF		_CLK
		;6x 8 Bit wurden gelesen und stehen in D1H/M/L, D2H/M/L
		;Bit 7 von DPos auf 1 setzen, zum Signalisieren, dass ein groer Frame gesendet wird, dann DPos und das erste Messdouble ber UART versenden
		;Die Sache mit den Mess-Doubles: Insgesamt mssen 1 Posbyte und 6 Messbyte bertragen werden, das dauert bei 38,4kBaud aber mind. 2ms
		;D.h. man wrde die nchsten Pos-Samples (1ms-Takt) verpassen
		;Deswegen werden immer nur 2 Messbytes zusammen mit einem Pos-Byte verschickt, das aber ber 3 Pos-Byte-Zyklen
		;Jede Sendung dauert dann nur 0,8ms, d.h. man verpasst nichts 
		BSF		DPos,7
		MOVF	DPos,_W
		CALL	RStx
		MOVF	D1H,_W
		CALL	RStx
		MOVF	D2H,_W
		CALL	RStx
		MOVLW	2			;Noch zwei Mess-Doubles mssen gesendet werden
		MOVWF	scnt
		BSF		_GIE		;Interrupts sind wieder erlaubt
		GOTO 	rnlp		;Bytes sind verschickt, auf nchsten Timer-Tick warten

SPos	;Es gibt keine HX711 Daten, aber evtl mssen noch ausstehende Mess-Doubles mitverschickt werden
		MOVF	scnt,_W		;scnt prfen
		BTFSS	_ZERO		;=0?
		GOTO	SDbl1		;Nein: Es muss noch ein Mess-Double mitverschickt werden
		BCF		DPos,7		;Ja: Bit 7 von DPos auf 0 setzen, zum Signalisieren, dass nur das DPos-Byte gesendet wird, dann senden
		MOVF	DPos,_W
		CALL 	RStx
		BSF		_GIE		;Interrupts sind wieder erlaubt
		GOTO 	rnlp		;Byte ist verschickt, auf nchsten Timer-Tick warten
SDbl1	BSF		DPos,7		;Es muss noch ein Mess-Double mitverschickt werden, deswegen Bit7=1
		MOVF	DPos,_W		
		CALL 	RStx
		DECFSZ	scnt,_R		;scnt-1, scnt=0?
		GOTO	SDbl2		;Nein: scnt war vorher 2, d.h. Mess-Double der Mid-Bytes verschicken
		MOVF	D1L,_W		;Ja: scnt war vorher 1, d.h. Mess-Double der Low-Bytes verschicken
		CALL 	RStx
		MOVF	D2L,_W
		CALL	RStx
		BSF		_GIE		;Interrupts sind wieder erlaubt
		GOTO 	rnlp		;Byte ist verschickt, auf nchsten Timer-Tick warten	
SDbl2	MOVF	D1M,_W		;Mess-Double der Mid-Bytes verschicken
		CALL 	RStx
		MOVF	D2M,_W
		CALL	RStx
		BSF		_GIE		;Interrupts sind wieder erlaubt
		GOTO 	rnlp		;Byte ist verschickt, auf nchsten Timer-Tick warten		

;;******* RStx *******************************************************
;Funktion: Sendet das Byte in W ber RS232 Protokoll(negative Logik!)
;Startbit(Low=5V), 8 Datenbits(Low=5V, High=0V) LSB first, Stopbit(High=0V) // Baudrate je nach delay in baud-Routine, jeder Wechsel zwischen zwei Bits dauert 10 Zyklen + 1*baud, No Parity
;Ein-/Ausgangswerte: W beinhaltet zu sendendes Byte
;Bentigte Subroutinen: baud
;Bentigte Variablen: temp, n, m
;Bentigte Register: keine
;Bentigte Defines: _RSTX, _CARRY (Transmit-Pin, Carry-Bit in STATUS)
;Unverndert bleibt: keine
;Verndert werden  : temp, n, m
RStx	MOVWF	temp			;W nach temp
		MOVLW 	.8
		MOVWF 	n				;8 Bits senden
		BSF   	_RSTX			;Startbit setzen (Low = 5V)
		NOP
		NOP
		NOP
		NOP
		NOP
		CALL  	baud			;Baudzeit warten
		RRF   	temp,1			;Bits durchschieben
		BTFSS 	_CARRY			;Bit checken = 1?
		GOTO	$+4				;nein: Sende 0
		NOP						;ja: Sende 1
		BCF	  	_RSTX			;High=0V
		GOTO	$+4				;Ab zur Pause 
		BSF   	_RSTX			;Low=5V
		NOP
		NOP
		CALL  	baud			;Baudzeit warten
		DECFSZ 	n,1				;n=n-1, n=0 (Byte zu Ende)?
		GOTO  	$-11			;nein: weiter mit naechstem Bit
		NOP						;ja: Stopbit setzen
		NOP
		NOP
		NOP
		NOP
		BCF   	_RSTX			;Stopbit setzen (High = 0V)
		CALL  	baud
		CALL	baud
		RETURN


;;******* RSrx *******************************************************
;Funktion: Wartet auf ein Startbit und liest dann ein Byte ber RS-232 Protokoll nach W (negative Logik! - ohne externe Inverter)
;Startbit(Low=5V), 8 Datenbits(Low=5V, High=0V) LSB first, Stopbit(High=0V) // Baudrate je nach delay in baud-Routine, jeder Wechsel zwischen zwei Bits dauert 10 Zyklen + 1*baud, No Parity
;Ein-/Ausgangswerte: W beinhaltet empfangenes Byte
;Bentigte Subroutinen: baud, baud2
;Bentigte Variablen: temp, n, m
;Bentigte Register: keine
;Bentigte Defines: _RSRX, _CARRY (Receive-Pin, Carry-Bit in STATUS)
;Unverndert bleibt: keine
;Verndert werden  : temp, n, m		
RSrx	CLRF	temp
		MOVLW	.8
		MOVWF	n
;		CALL	baud2		;Startbit wurde registriert, warte 1/2 baudzeit (danach gleich noch eine ganze) = 1,5 baud seit der Startbitflanke = Mitte vom ersten bit
		CALL	baud		;Warte volle baudrate-10 Zyklen (werden in Schleife verbraten), um in naechstes Bit zu kommen
		NOP
		BTFSC	_RSRX		;Pruefe Bit
		GOTO	$+3			;=1: (=Logisch 0)
		BSF		_CARRY		;=0: (=Logisch 1) Carry mit 1 besetzen
		GOTO	$+3
		BCF		_CARRY		;Carry mit 0 besetzen
		NOP					;Damit gleich lang wie BCF-Zweig
		RRF		temp,1		;Carry in Byte verschieben (RS232:LSB first -> Bits werden von links reingeschoben)
		DECFSZ	n,1
		GOTO	$-10
		CALL	baud		;Sicherstellen, dass Stopbit vor nchstem Receive empfangen wird
		MOVF	temp,0		;gelesenes Byte nach W
		RETURN
	

;******** baud **********************
;Funktion: Warteschleife fuer Baudrate 38400 (=26us) bei 4MHz/4 Takt (=26 Zyklen). 
;10 Zyklen werden schon in der RS232 Routine zwischen jedes Bit gesetzt d.h. zusammen mit aufrufendem CALL werden hier 16 Zyklen verbraten
;2+2+(3*3)+1+2 (CALL + 2*MOV + 3*Schleife + NOP + RETURN)
;Ein-/Ausgangswerte: keine
;Bentigte Subroutinen: keine
;Bentigte Variablen: m
;Bentigte Register: keine
;Bentigte Defines: keine
;Unverndert bleibt: keine
;Verndert werden  : m	
baud	MOVLW  	.3	
		MOVWF  	m
		DECFSZ 	m,1			
		GOTO  	$-1
		NOP
		NOP
		RETURN


;******** baud2 **********************
;Funktion: Warteschleife fuer 1/2 Baudrate 38400 (=13us) bei 4MHz/4 Takt (=13 Zyklen). 
;Es werden keine Zwischentakte bercksichtigt d.h. zusammen mit aufrufendem CALL werden hier 13 Zyklen verbraten
;2+2+(2*3)+2+2 (CALL + 2*MOV + 2*Schleife +2NOP +  RETURN)
;Ein-/Ausgangswerte: keine
;Bentigte Subroutinen: keine
;Bentigte Variablen: m
;Bentigte Register: keine
;Bentigte Defines: keine
;Unverndert bleibt: keine
;Verndert werden  : m	
baud2	MOVLW  	.2	
		MOVWF  	m
		DECFSZ 	m,1			
		GOTO  	$-1
		NOP
		NOP
		RETURN
			
		
		END
